#!/usr/bin/env python
# Copyright Salient Predictions 2024

"""Define the Location class."""

# Usage example:
# python location.py --lat 42 --lon -73

import os

import cartopy.crs as ccrs
import cartopy.feature as cfeature
import geopandas as gpd
import matplotlib.pyplot as plt
import pandas as pd

from . import constants, login_api

# location_file supports a variety of potential column names:
# fmt: off
LON_NAMES = ["lon", "Lon", "LON", "longitude", "Longitude", "LONGITUDE", "lng", "x", "X"]
LAT_NAMES = ["lat", "Lat", "LAT", "latitude", "Latitude", "LATITUDE", "y", "Y"]
NAM_NAMES = ["Location", "location", "loc", "id", "ID", "Place", "place", "Coordinates", "coordinates", "Name", "name"]
# fmt:on


class Location:
    """Define geographical parameters for an API query.

    The Salient API defines location by a single latitude/longitude,
    multiple latitude/longitude pairs in a single location_file, or a polygon
    defined in a shapefile.  This class manages these options.
    """

    def __init__(
        self,
        lat: float | None = None,
        lon: float | None = None,
        location_file: str | list[str] | None = None,
        shapefile: str | list[str] | None = None,
        region: str | list[str] | None = None,
    ):
        """Initialize a Location object.

        Only one of the following 4 options should be used at a time: lat/lon,
        location_file, shapefile, or region.

        Args:
            lat (float): Latitude in degrees, -90 to 90.
            lon (float): Longitude in degrees, -180 to 180.
            location_file (str | list[str]): Path(s) to CSV file(s) with latitude and longitude columns.
            shapefile (str | list[str]): Path(s) to a shapefile(s) with a polygon defining the location.
            region (str | list[str]): Accepts continents, countries, or U.S. states (e.g. "usa")
                Only available for `hindcast_summary()`
        """
        self.lat = lat
        self.lon = lon
        self.location_file = self._expand_user_files(location_file)
        self.shapefile = self._expand_user_files(shapefile)
        self.region = region

        self._validate()

    @staticmethod
    def _expand_user_files(files: str | list[str] | None) -> str | list[str] | None:
        """Strip directory names from location_file and shapefile.

        When referencing user files, the user may specify them with
        a directory name because that's how functions like upload_*
        create them.  But the API doesn't have a directory structure and
        only uses the file name.  So we need to strip off the directory.

        Also, handles vectorization of file names in case user passes in
        a list or a comma-separated string.

        :exclude:
        """
        files = constants._expand_comma(files)

        if files is None:
            pass
        elif isinstance(files, str):
            files = os.path.basename(files)
        elif isinstance(files, list):
            files = [os.path.basename(f) for f in files]

        return files

    def asdict(self, **kwargs) -> dict:
        """Render as a dictionary.

        Generates a dictionary representation that can be encoded into a URL.
        Will contain one and only one location_file, shapefile, or lat/lon pair.

        Args:
            **kwargs: Additional key-value pairs to include in the dictionary.
                Will validate some common arguments that are shared across API calls.
        """
        if self.location_file:
            dct = {"location_file": self.location_file, **kwargs}
        elif self.shapefile:
            dct = {"shapefile": self.shapefile, **kwargs}
        elif self.region:
            dct = {"region": self.region, **kwargs}
        else:
            dct = {"lat": self.lat, "lon": self.lon, **kwargs}

        if "apikey" in dct and dct["apikey"] is not None:
            dct["apikey"] = login_api._get_api_key(dct["apikey"])

        if "start" in dct:
            dct["start"] = constants._validate_date(dct["start"])
        if "end" in dct:
            dct["end"] = constants._validate_date(dct["end"])
        if "forecast_date" in dct:
            dct["forecast_date"] = constants._validate_date(dct["forecast_date"])

        if "version" in dct:
            dct["version"] = constants._expand_comma(
                val=dct["version"],
                valid=constants.MODEL_VERSIONS,
                name="version",
                default=constants.get_model_version(),
            )

        if "shapefile" in dct and "debias" in dct and dct["debias"]:
            raise ValueError("Cannot debias with shapefile locations")

        return dct

    def load_location_file(self, destination: str = "-default") -> gpd.GeoDataFrame:
        """Load the location file(s) into a DataFrame.

        Args:
            destination (str): The directory where the file is located.
                Defaults to the default directory via `get_file_destination`.

        Returns:
            gpd.GeoDataFrame: The location data in a DataFrame.  If multiple files are loaded,
                the DataFrames will be concatenated into a single DataFrame with an
                additional column `file_name` that documents the source file.

                location_file suports a variety of column names for latitude and longitude.
                This method standardizes them to `lat` and `lon`.
        """
        assert self.location_file is not None

        destination = constants.get_file_destination(destination)

        if isinstance(self.location_file, str):
            geo = pd.read_csv(os.path.join(destination, self.location_file))
        else:
            # location files may be a list of files.  Load all of them.
            geo = pd.concat(
                [
                    pd.read_csv(os.path.join(destination, f)).assign(file_name=f)
                    for f in self.location_file
                ]
            )

        return self._as_geoframe(geo)

    @staticmethod
    def _as_geoframe(geo: pd.DataFrame) -> gpd.GeoDataFrame:
        """Normalize column names on a DataFrame and convert to a GeoDataFrame.

        Regardless of the column names in the input DataFrame, this method
        will return a GeoDataFrame with columns `lat`,`lon`, and `name`

        :exclude:
        """
        col_names = set(geo.columns)
        lon_name = next((name for name in LON_NAMES if name in col_names), None)
        lat_name = next((name for name in LAT_NAMES if name in col_names), None)
        nam_name = next((name for name in NAM_NAMES if name in col_names), None)

        assert lon_name is not None, f"Missing longitude column in {col_names}"
        assert lat_name is not None, f"Missing latitude column in {col_names}"

        geo.rename(columns={lon_name: "lon", lat_name: "lat"}, inplace=True)

        if nam_name is not None:
            # "name" column is optional and not guaranteed to be present.
            geo.rename(columns={nam_name: "name"}, inplace=True)

        geo = gpd.GeoDataFrame(geo, geometry=gpd.points_from_xy(geo.lon, geo.lat))

        return geo

    def _validate(self):
        if self.location_file:
            assert not self.lat, "Cannot specify both lat and location_file"
            assert not self.lon, "Cannot specify both lon and location_file"
            assert not self.region, "Cannot specify both region and location_file"
            assert not self.shapefile, "Cannot specify both shape_file and location_file"
        elif self.shapefile:
            assert not self.region, "Cannot specify both region and shapefile"
            assert not self.lat, "Cannot specify both lat and shape_file"
            assert not self.lon, "Cannot specify both lon and shape_file"
        elif self.region:
            assert not self.lat, "Cannot specify both lat and region"
            assert not self.lon, "Cannot specify both lon and region"
            assert not self.location_file, "Cannot specify both location_file and region"
            assert not self.shapefile, "Cannot specify both shape_file and region"
        else:
            assert self.lat, "Must specify lat & lon, location_file, shapefile, or region"
            assert self.lon, "Must specify lat & lon, location_file, shapefile, or region"
            assert -90 <= self.lat <= 90, "Latitude must be between -90 and 90 degrees"
            assert -180 <= self.lon <= 180, "Longitude must be between -180 and 180 degrees"

    def plot_locations(self, title: str = None, weight: str = None, pad: float = 1.0):
        """Show location points on a map.

        Args:
            title (str): The title of the plot.
            weight (str): The column name in the location file to use to weight the points.
            pad (float): Adds extent to the lat/lon bounding boxon the map.
        """
        geo = self.load_location_file()

        if title is None:
            title = str(self.location_file)

        if weight is not None:
            weight = geo[weight] * 10

        fig, ax = plt.subplots(figsize=(6, 6), subplot_kw={"projection": ccrs.Mercator()})
        min_lon, max_lon = geo["lon"].min() - pad, geo["lon"].max() + pad
        min_lat, max_lat = geo["lat"].min() - pad, geo["lat"].max() + pad

        ax.set_extent([min_lon, max_lon, min_lat, max_lat], crs=ccrs.PlateCarree())
        ax.add_feature(cfeature.BORDERS)
        # ax.add_feature(cfeature.COASTLINE)
        ax.add_feature(cfeature.LAND, edgecolor="black")
        ax.add_feature(cfeature.OCEAN, facecolor="aqua")
        ax.add_feature(cfeature.LAKES, facecolor="aqua")

        ax.scatter(
            geo["lon"],
            geo["lat"],
            s=weight,
            color="dodgerblue",
            alpha=0.5,
            transform=ccrs.PlateCarree(),
        )

        for lon, lat, name in zip(geo["lon"], geo["lat"], geo["name"]):
            ax.text(
                lon,
                lat,
                name,
                fontsize=8,
                ha="center",
                va="center",
                transform=ccrs.PlateCarree(),
            )

        plt.title(title)
        plt.show()

        return fig, ax

    def __str__(self):
        """Return a string representation of the Location object.

        :exclude:
        """
        if self.location_file:
            return f"location file: {self.location_file}"
        elif self.shapefile:
            return f"shape file: {self.shapefile}"
        elif self.region:
            return f"region: {self.region}"
        else:
            return f"({self.lat}, {self.lon})"

    def __eq__(self, other):
        """Check if two Location objects are equal.

        :exclude:
        """
        if self.location_file:
            return self.location_file == other.location_file
        elif self.shapefile:
            return self.shapefile == other.shape_file
        elif self.region:
            return self.region == other.region
        else:
            return self.lat == other.lat and self.lon == other.lon

    def __ne__(self, other):
        """Check if two Location objects are not equal.

        :exclude:
        """
        return not self.__eq__(other)



#!/usr/bin/env python
# Copyright Salient Predictions 2024

"""Skill and validation."""

import numpy as np
import pandas as pd
import xarray as xr


def crpss(
    forecast: xr.Dataset | xr.DataArray,
    reference: xr.Dataset | xr.DataArray,
) -> xr.Dataset | xr.DataArray:
    """Continuous Ranked Probability Skill Score.

    CRPSS measures the relative skill improvement of one CRPS forecast over another.
    Positive values can roughly interpreted as percentages, so a CRPSS of 0.10 means a 10% improvement.
    Zero values mean no improvement, and negative values mean the forecast is worse than the reference.

    Args:
        forecast: The forecast data
        reference: The reference baseline data

    Returns:
        xr.DataArray: The CRPSS result
    """
    skill_score = 1 - (forecast / reference)

    skill_score = _set_array_name(skill_score, "crpss")

    return skill_score


def crps(
    observations: xr.Dataset | xr.DataArray | str,
    forecasts: xr.Dataset | xr.DataArray | str | list[str] | pd.DataFrame,
    qnt_dim: str = "quantiles",
) -> xr.DataArray:
    """Calculate Continuous Ranked Probability Score.

    CRPS is used to calculate the skill of a probabilistic forecast.
    CRPS is defined as 2 times the integral of the quantile loss over the distribution.
    Zero CRPS indicates a perfect forecast.

    Args:
        observations: `DataArray` of observed values, aka "truth", or a `nc` filename.
            `observations` should usually have the same coordinates as `forecasts`.
            If `observations` has a `daily` timescale to match
            a `weekly`, `monthly`, or `quarterly` timescale for `forecasts.`
        forecasts: `DataArray` of forecasted values.  May also be:
          * a file reference to a `nc` file
          * a vector of file references to `nc` files
          * a DataFrame with a `file_name` column
        qnt_dim: Name of the quantile dimension in the forecast array.
            Defaults to `quantiles`, which is the dimension name returned by
            `forecast_timeseries`.

    Returns:
        xr.DataArray: The CRPS of the `forecasts` quantiles vs. the `observations`.

    """
    return _calc_skill(observations, forecasts, skill_func=_crps_core, qnt_dim=qnt_dim)


def _crps_core(
    observations: xr.Dataset | xr.DataArray,
    forecasts: xr.Dataset | xr.DataArray,
    qnt_dim="quantiles",
) -> xr.Dataset | xr.DataArray:
    """The fundamental Continuous Ranked Probability Score calculation.

    This function differs from the user-visible `crps` function in that it does not
    perform any vectorization, coordinate alignment, or validation.
    """
    diff = observations - forecasts
    qnt_val = diff[qnt_dim]

    qnt_score = 2 * np.maximum(qnt_val * diff, (qnt_val - 1) * diff)
    skill = qnt_score.integrate(coord=qnt_dim)
    skill.attrs = forecasts.attrs.copy()

    skill = _set_array_name(skill, "crps")

    return skill


def _calc_skill(
    observations: xr.Dataset | xr.DataArray | str,
    forecasts: xr.Dataset | xr.DataArray | str | list[str] | pd.DataFrame,
    skill_func: callable,
    **kwargs,
) -> xr.DataArray:
    """Calculate a skill score.

    Handles vectorization and coordinate alignment for skill functions.  Skill itself
    is calculated by the `skill_func` function.

    Args:
        observations: `DataArray` or `Dataset` of observed ("truth") values , or a `nc` filename.
            `observations` should usually have the same coordinates as `forecasts`.
            If `observations` has a `daily` timescale, the system will aggregate match
            a `weekly`, `monthly`, or `quarterly` timescale for `forecasts.`
        forecasts: `Dataset` or `DataArray` of forecasted values.  May also be:
          * a file reference to a `nc` file
          * a vector of file references to `nc` files, which will calculate skill for
            each file and return the average across all forecasts.
          * a DataFrame with a `file_name` column
        skill_func: The skill function to use.
        **kwargs: Additional arguments to pass to `skill_func`.

    Returns:
        xr.DataArray: The skill of the `forecast` values vs. the `observation`s.

    """

    def extract_df_files(obj, col_name="file_name"):
        """Extract file names from a DataFrame, if passed in."""
        if isinstance(obj, pd.DataFrame):
            assert col_name in forecasts.columns, f"DataFrame must have a '{col_name}' column."
            obj = obj[col_name].tolist()
        return obj

    if observations is None or observations is pd.NA:
        return None
    elif isinstance(observations, str):
        observations = xr.load_dataset(observations)
    elif not isinstance(observations, xr.DataArray) and not isinstance(observations, xr.Dataset):
        raise ValueError(
            f"observations {type(observations)} must be Dataset, DataArray, or filename"
        )

    forecasts = extract_df_files(forecasts)
    if forecasts is None or forecasts is pd.NA:
        return None
    elif isinstance(forecasts, str):
        # We want to load_datasaet instead of load_dataarray in order to preserve forecast_period
        forecasts = xr.load_dataset(forecasts)
    elif isinstance(forecasts, list):
        skill = [_calc_skill(observations, fcst, skill_func, **kwargs) for fcst in forecasts]
        skill = [s for s in skill if s is not None]
        if len(skill) == 0:
            return None
        elif len(skill) == 1:
            return skill[0]

        # What if forecast=all and there are multiple forecast dates?
        concat_dim = _find_coord(skill[0], "forecast_date", strict=True)
        skill = xr.concat(skill, dim=concat_dim)

        skill = skill.mean(dim=concat_dim, keep_attrs=True)

        return skill
    elif not isinstance(forecasts, xr.DataArray) and not isinstance(forecasts, xr.Dataset):
        raise ValueError(
            f"forecast {type(forecasts)} must be a Dataset, DataArray, DataFrame[file_name], or filename."
        )

    # At this point, all vectorization and file loading should be done.  Now we need to make sure
    # that the coordinates match.
    if "time" in observations.coords:
        lead_dim = _find_coord(forecasts, "lead_", strict=False)
        if lead_dim is not None:
            observations = align_daily_obs_to_lead(observations, forecasts, lead_dim[5:])
        elif "time" in forecasts.coords:
            (observations, forecasts) = _align_time_to_lead(observations, forecasts)

    return skill_func(observations, forecasts, **kwargs)


def _set_array_name(ds: xr.Dataset | xr.DataArray, name: str) -> xr.Dataset | xr.DataArray:
    """Set the name of the data array or first dataset variable."""
    if isinstance(ds, xr.DataArray):
        ds.name = name
    elif isinstance(ds, xr.Dataset):
        if len(ds.data_vars) == 1:
            ds = ds.rename_vars({list(ds.data_vars)[0]: name})

    ds.attrs["short_name"] = name
    ds.attrs["long_name"] = (
        ds.attrs["long_name"] + " " if "long_name" in ds.attrs else ""
    ) + name.upper()

    return ds


def _extract_array_name(ds: xr.Dataset | xr.DataArray) -> str:
    """Extract the name of the first data variable in a dataset."""
    if isinstance(ds, xr.Dataset):
        # get the string name of the first forecast data var:
        return next(iter(ds.data_vars))
    elif isinstance(ds, xr.DataArray):
        return ds.name
    else:
        raise ValueError(f"Must be a Dataset or DataArray, not {type(ds)}")


def align_daily_obs_to_lead(
    observations: xr.Dataset | xr.DataArray,
    forecasts: xr.Dataset | xr.DataArray,
    timescale: str,
) -> xr.Dataset | xr.DataArray:
    """Convert daily observations to match forecasts denominated by a coarse lead time.

    Args:
        observations: Daily observed values with coordinate `time`.
        forecasts: The forecasted values with coordinate `lead_<timescale>` and
          `forecast_period_<timescale>`.
        timescale: The forecast period of the forecast, corresponding to data variables
            `lead_<timescale>` and `forecast_period_<timescale>`.
            Will typically be `weekly`, `monthly`, or `quarterly`.

    Returns:
        xr.Dataset | xr.DataArray: An aggregated version of `observations` with coordinate
            `lead_<timescale>` instead of `time`.
    """
    lead_name = f"lead_{timescale}"
    data_name = _extract_array_name(forecasts)
    period_name = f"forecast_period_{timescale}"

    lead_vals = forecasts[lead_name].values

    # groupby_bins creates bins that are right-inclusive, so we start the
    # binning one day early:
    first_day = forecasts[period_name].isel(nbnds=0)[0] - np.timedelta64(1, "D")
    bins = np.append(first_day, forecasts[period_name].isel(nbnds=1))

    observations = (
        observations.groupby_bins("time", bins)
        .mean()
        .assign_coords(lead=("time_bins", lead_vals))
        .rename({"lead": lead_name})
        .swap_dims({"time_bins": lead_name})
        .drop_vars("time_bins")
    )

    if isinstance(observations, xr.Dataset):
        obs_name = _extract_array_name(observations)
        observations = observations.rename({obs_name: data_name})
    elif isinstance(observations, xr.DataArray):
        observations.name = data_name

    return observations


def _find_coord(
    ds: xr.DataArray | xr.Dataset, starts_with: str, strict: bool = True
) -> str | None:
    """Find a coordinate that starts with a given string.

    Args:
        ds: The dataset or dataarray to search
        starts_with: Search for coordinates that begin with this string
        strict: If True, raise an error if no match is found.

    Returns:
        str: The name of the coordinate that starts with `starts_with`
            or `None` if no coordinate was found
    """
    found = next((coord for coord in ds.coords if coord.startswith(starts_with)), None)
    if strict:
        assert found is not None, f"No {starts_with} coordinate found."
    return found


def _align_time_to_lead(
    observations: xr.DataArray, forecasts: xr.DataArray
) -> (xr.DataArray, xr.DataArray):
    """Align or aggregate observations to match the granularity of forecasts.

    Args:
        observations (xr.DataArray): The observed values with coordinate `time`.
        forecasts (xr.DataArray): The forecasted values with coordinate `time`.

    Returns:
        tuple: Aligned or aggregated observations and forecasts with updated 'lead' coordinates.
    """
    # Convert time coordinates to pd.DatetimeIndex and infer frequencies

    # obs_time = _ensure_freq(observations.time.values)
    fcst_time = _ensure_freq(forecasts.time.values)

    lead_times = range(1, len(forecasts.time) + 1)
    forecasts = forecasts.assign_coords(lead=("time", lead_times))

    # TODO: trim observations to match forecast time range for efficiency.
    observations = observations.resample(time=fcst_time.freq).mean()

    observations, forecasts = xr.align(observations, forecasts, join="inner")
    observations = observations.assign_coords(lead=("time", forecasts.lead.values))

    forecasts = forecasts.swap_dims({"time": "lead"})
    observations = observations.swap_dims({"time": "lead"})

    return observations, forecasts


def _ensure_freq(time) -> pd.DatetimeIndex:
    """Ensure that a time vector has a 'freq' attribute for resampling."""
    if not isinstance(time, pd.DatetimeIndex):
        time = pd.to_datetime(time)

    if time.freq is not None:
        # Frequency is already set
        return time
    elif len(time) < 2:
        # Not enough data to infer or calculate frequency
        raise ValueError("Not enough data points to determine frequency.")
    elif len(time) == 2:
        # Manually calculate frequency from two dates.
        step = time[1] - time[0]
        days = step.days
        if days == 1:
            freq = "D"
        elif days == 7:
            day_of_week = time[0].day_name()
            freq = f"W-{day_of_week[:3].upper()}"
        elif 28 <= days <= 31:
            freq = "MS"  # ms= momth start
        elif 89 <= days <= 92:
            # Get the month for the first date
            month_of_year = time[0].month
            # Map month to the corresponding quarter start
            if month_of_year in [1, 2, 3]:
                freq = "Q-JAN"
            elif month_of_year in [4, 5, 6]:
                freq = "Q-APR"
            elif month_of_year in [7, 8, 9]:
                freq = "Q-JUL"
            elif month_of_year in [10, 11, 12]:
                freq = "Q-OCT"
        else:
            raise ValueError(f"Unknown time step: {days} days.")
    else:
        # Infer frequency from three or more dates
        freq = pd.infer_freq(time)
        if freq is None:
            raise ValueError("Unable to infer frequency from data.")

    return pd.DatetimeIndex(time, freq=freq)
